/*
	this file is under GPL
*/

#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include <tchar.h>
#include <wchar.h>
#define KMELEON_PLUGIN_EXPORTS
#include <kmeleon_plugin.h>

#include <nsCOMPtr.h>
#include <nsIWebBrowser.h>
#include <nsIWebBrowserFocus.h>
#include <nsIDOMElement.h>
#include <nsIDOMNSEditableElement.h>
#include <nsIDOMRange.h>
#include <nsIEditor.h>
#include <nsISelection.h>
#include <nsIInlineSpellChecker.h>
#include <nsIEditorSpellCheck.h>
#include <nsStringAPI.h>
#include <nsEmbedString.h>
#include <nsMemory.h>

#include <nsXPCOMCID.h>
#include <nsIProperties.h>
#include <nsILocalFile.h>

#include <nsIServiceManager.h>
#include <nsIAccessible.h>
#include <nsIAccessibleRetrieval.h>
#include <ISimpleDOMText.h>
#include <oleacc.h>
#include <servprov.h>

#include <algorithm>


#define PLUGIN_NAME "Spellcheck test plugin"

#define LABEL_TITLE "K-Meleon spellchecker"
#define LABEL_ADDWORD "Add this word"
#define LABEL_IGNOREWORD "Ignore this word"
#define LABEL_SELECTDIC "Dictionaries"
#define LABEL_SETDIC "Set Default"
#define LABEL_ENABLE "Spellchecker is disabled.\r\nEnable?"
#define LABEL_DISABLE "Disable spellchecker"

#define L_ID_DISABLE 1
#define L_ID_ADDWORD 2
#define L_ID_IGNOREWORD 3
#define L_ID_SETDIC 4

#define PREFKEY_DICTIONARY     "spellchecker.dictionary"
#define PREFKEY_DICTIONARY_DIR "kmeleon.plugins.spelltest.dictDir"
#define PREFKEY_MENU_POSITION  "kmeleon.plugins.spelltest.position"
#define PREFKEY_SUGGEST_ORDER  "kmeleon.plugins.spelltest.order"
#define PREFKEY_MAX_SUGGESTS   "kmeleon.plugins.spelltest.max_suggests"

long DoMessage(const char *, const char *, const char *, long, long);
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
void Load();
void Create(HWND);
int DoAccel(char *);
BOOL DoCommand(HWND, BOOL);
BOOL SetDictionariesDir();

WNDPROC KMeleonWndProc;
UINT g_id;

kmeleonPlugin kPlugin = {
	KMEL_PLUGIN_VER,
	PLUGIN_NAME,
	DoMessage
};

long DoMessage(const char *to, const char *from, const char *subject,
               long data1, long data2)
{
	if (to[0] == '*' || stricmp(to, kPlugin.dllname) == 0) {
		if (stricmp(subject, "Load") == 0) {
			Load();
		}
		else if (stricmp(subject, "Create") == 0) {
			Create((HWND) data1);
		}
		else if (stricmp(subject, "DoAccel") == 0) {
			int *id = (int *) data2;
			*id = DoAccel((char *) data1);
		}
		else return 0;
		return 1;
	}
	return 0;
}


void Load()
{
	SetDictionariesDir();
}


void Create(HWND hWndParent)
{
	g_id = kPlugin.kFuncs->GetCommandIDs(2);
	KMeleonWndProc = (WNDPROC) GetWindowLong(hWndParent, GWL_WNDPROC);
	SetWindowLong(hWndParent, GWL_WNDPROC, (LONG)WndProc);
}


int DoAccel(char *param)
{
	if (lstrcmpi(param, _T("here")) == 0 || lstrcmpi(param, _T("word")) == 0) {
		return g_id + 1;
	}
	else if (lstrcmpi(param, _T("mouse")) == 0) {
		return g_id;
	}
	int pos = 0;
	kPlugin.kFuncs->GetPreference(PREF_INT, PREFKEY_MENU_POSITION, &pos, &pos);
	if (pos) {
		return g_id + 1;
	}
	return g_id;
}


LRESULT CALLBACK WndProc(HWND hWnd, UINT message,
	WPARAM wParam, LPARAM lParam)
{
	switch (message) {
	case WM_COMMAND:
		if (LOWORD(wParam) == g_id) {
			DoCommand(hWnd, FALSE);
		}
		else if (LOWORD(wParam) == g_id + 1) {
			DoCommand(hWnd, TRUE);
		}
		break;
	}
	return CallWindowProc(KMeleonWndProc, hWnd, message, wParam, lParam);
}


extern "C" {
	BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
	{
		switch (dwReason) {
		case DLL_PROCESS_DETACH:
			break;
		}
		return TRUE;
	}
	
	KMELEON_PLUGIN kmeleonPlugin *GetKmeleonPlugin() {
		return &kPlugin;
	}
	
	KMELEON_PLUGIN int DrawBitmap(DRAWITEMSTRUCT *dis) {
		return 14; // 14 = icon width
	}
}


/*
  K-Meleon is built by Visual C++, then C++ classes in DLLs may not be use
  other compilers because of different name mangling rule.
  And I use small alternative class.
  If you remake this by VC, you should use usual mozilla classes.
 */


/*
  tiny alternative of nsCOMPtr
 */
template <class T>
struct miniCOMPtr {
	miniCOMPtr() : m_ptr(nsnull) {}
	~miniCOMPtr() { NS_IF_RELEASE(m_ptr); }
	operator bool() { return m_ptr; }
	operator T* () { return m_ptr; }
	T** operator&() { return &m_ptr; }
	T* operator->() { return m_ptr; }
	nsresult doQueryInterface(nsISupports* i);
	nsresult doGetService(const char* cid);
	template <class U> nsresult QueryInterface(miniCOMPtr<U>& i) {
		return i.doQueryInterface(m_ptr);
	}
private:
	T* m_ptr;
};

template <class T>
nsresult miniCOMPtr<T>::doQueryInterface(nsISupports* i) {
	nsresult rv;
	void* ptr;
	rv = i->QueryInterface(NS_GET_IID(T), &ptr);
	if (NS_SUCCEEDED(rv)) {
		NS_IF_RELEASE(m_ptr);
		m_ptr = reinterpret_cast<T*>(ptr);
	}
	return rv;
}

nsresult GetServiceByContractID(const char* cid, const nsIID& iid, void** result)
{
	nsresult rv;
	miniCOMPtr<nsIServiceManager> servicemanager;
	rv = NS_GetServiceManager(&servicemanager);
	NS_ENSURE_SUCCESS(rv, rv);
	NS_ENSURE_TRUE(servicemanager, NS_ERROR_FAILURE);
	return servicemanager->GetServiceByContractID(cid, iid, result);
}

template <class T>
nsresult miniCOMPtr<T>::doGetService(const char* cid)
{
	nsresult rv;
	void* ptr;
	rv = GetServiceByContractID(cid, NS_GET_IID(T), &ptr);
	if (NS_SUCCEEDED(rv)) {
		NS_IF_RELEASE(m_ptr);
		m_ptr = reinterpret_cast<T*>(ptr);
	}
	return rv;
}


/*
  Set dictionaries path.
  
  seamonkey/extensions/spellcheck/idl/mozISpellCheckingEngine.idl
  seamonkey/extensions/spellcheck/hunspell/src/mozHunspell.cpp
 */
BOOL SetDictionariesDir()
{
	nsresult rv;
	PRUnichar dirname[MAX_PATH] = {0};
	long len;
	
	len = kPlugin.kFuncs->GetPreference(PREF_UNISTRING, PREFKEY_DICTIONARY_DIR, NULL, (void *) L"");
	if (len > MAX_PATH) return FALSE;
	kPlugin.kFuncs->GetPreference(PREF_UNISTRING, PREFKEY_DICTIONARY_DIR, dirname, (void *) L"");
	if (! wcslen(dirname)) return TRUE;
	
	miniCOMPtr<nsIProperties> dirsvc;
	rv = dirsvc.doGetService(NS_DIRECTORY_SERVICE_CONTRACTID);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(dirsvc, FALSE);
	
	miniCOMPtr<nsILocalFile> directory;
	nsEmbedString dirstr(dirname);
	rv = NS_NewLocalFile(dirstr, PR_FALSE, &directory);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(directory, FALSE);
	
	// "DictD" is DICTIONARY_SEARCH_DIRECTORY
	// defined in mozISpellCheckingEngine.idl
	rv = dirsvc->Set("DictD", directory);
	NS_ENSURE_SUCCESS(rv, FALSE);
	return TRUE;
}


/*
  Get nsIEditor on the focus from nsIWebBrowser.
 */
BOOL get_editor(HWND hwnd, miniCOMPtr<nsIEditor>& editor)
{
	nsresult rv;
	
	miniCOMPtr<nsIWebBrowser> browser;
	if (! kPlugin.kFuncs->GetMozillaWebBrowser(hwnd, &browser)) {
		return FALSE;
	}
	
	miniCOMPtr<nsIWebBrowserFocus> focus;
	rv = browser.QueryInterface(focus);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(focus, FALSE);
	
	miniCOMPtr<nsIDOMElement> elem;
	rv = focus->GetFocusedElement(&elem);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(elem, FALSE);
	
	miniCOMPtr<nsIDOMNSEditableElement> ee;
	rv = elem.QueryInterface(ee);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(ee, FALSE);
	
	rv = ee->GetEditor(&editor);
	NS_ENSURE_TRUE(editor, FALSE);
	
	return TRUE;
}


/*
  Minimal string list class
 */
struct miniWords
{
	/* construct with alloc */
	miniWords(int size);
	/* construct by pointer */
	miniWords(PRUnichar** p, int size)
		: m_length(size), m_size(size), m_words(p) {}
	~miniWords() { clear(); }
	int length() { return m_length; }
	int size() { return m_size; }
	void push(PRUnichar* p) {
		if (m_length < m_size) 
			m_words[m_length ++] = p;
	}
	void clear();
	PRUnichar* operator[] (const int i) {
		return (i >= 0 && i < m_length) ? m_words[i] : 0;
	}
	void sort();
private:
	static bool comp(PRUnichar *a, PRUnichar *b) {
		return wcscmp(a, b) < 0;
	}
	int m_length, m_size;
	PRUnichar** m_words;
};

miniWords::miniWords(int size) : m_length(0), m_size(0), m_words(0) {
	m_words = (PRUnichar**) nsMemory::Alloc(sizeof(PRUnichar*) * size);
	if (m_words) {
		m_size = size;
	}
}

void miniWords::clear()
{
	if (m_words) {
		for (int i = 0; i < m_length; ++ i) {
			nsMemory::Free(m_words[i]);
		}
		nsMemory::Free(m_words);
		m_words = 0;
		m_length = m_size = 0;
	}
}

void miniWords::sort()
{
	if (m_words) {
		std::sort(m_words, m_words + m_length, miniWords::comp);
	}
}

template <typename T>
struct miniPtr
{
	/* construct by pointer */
	miniPtr(T* p = 0) : m_ptr(p) {}
	~miniPtr() { if (m_ptr) nsMemory::Free(m_ptr); }
	operator T* () { return m_ptr; }
	T** operator&() { return &m_ptr; }
private:
	T* m_ptr;
};

/*
  UTF16 to ANSI converter
 */
struct miniMBString
{
	miniMBString(const PRUnichar* c);
	~miniMBString() { delete[] m_string; }
	operator char* () { return m_string; }
private:
	char* m_string;
};

miniMBString::miniMBString(const PRUnichar* c) : m_string(0) {
	static const DWORD flags = WC_COMPOSITECHECK;
	static const UINT cp = CP_ACP;
	if (c) {
		int length = WideCharToMultiByte(cp, flags, c, -1, NULL, 0, NULL, NULL);
		if (length > 0) {
			m_string = new char[length];
			if (m_string) {
				m_string[0] = '\0';
				WideCharToMultiByte(cp, flags, c, -1, m_string, length, NULL, NULL);
			}
		}
	}
}

/*
  ANSI to UTF16 converter
 */
struct miniWString
{
	miniWString(const char* c);
	~miniWString() { delete[] m_string; }
	operator LPWSTR () { return m_string; }
private:
	LPWSTR m_string;
};

miniWString::miniWString(const char* c) : m_string(0) {
	static const UINT cp = CP_ACP;
	static const DWORD flags = MB_PRECOMPOSED;
	if (c) {
		int length = MultiByteToWideChar(cp, flags, c, -1, NULL, 0);
		if (length > 0) {
			m_string = new WCHAR[length];
			if (m_string) {
				m_string[0] = '\0';
				MultiByteToWideChar(cp, flags, c, -1, m_string, length);
			}
		}
	}
}

BOOL append_menuA(HMENU hMenu, UINT flags, UINT item, PRUnichar* label)
{
	return AppendMenuA(hMenu, flags, item, miniMBString(label));
}
BOOL append_menuA_tr(HMENU hMenu, UINT flags, UINT item, char* label)
{
	return AppendMenuA(hMenu, flags, item, kPlugin.kFuncs->Translate(label));
}
BOOL append_menuW(HMENU hMenu, UINT flags, UINT item, PRUnichar* label)
{
	return AppendMenuW(hMenu, flags, item, label);
}
BOOL append_menuW_tr(HMENU hMenu, UINT flags, UINT item, char* label)
{
	return AppendMenuW(hMenu, flags, item, miniWString(kPlugin.kFuncs->Translate(label)));
}

struct miniHMenu
{
	miniHMenu() : m_hmenu(0) {};
	~miniHMenu() { if (m_hmenu) DestroyMenu(m_hmenu); }
	HMENU operator= (HMENU hmenu) { return m_hmenu = hmenu; }
	operator HMENU () { return m_hmenu; }
	HMENU release() { HMENU m = m_hmenu; m_hmenu = 0; return m; }
private:
	HMENU m_hmenu;
};


/*
  show popup menu
  
  returns
  -1 : error
   0 : no select
   1 : disable spellcheck
   2 : add word to personal dictionary
   3 : ignore word
   4 : set default dictionary
   0x1000 - 0x1fff : dictionaries
   0x2000 - 0x2fff : suggested words
 */
INT select_menu(HWND hwnd, miniWords& suggest, miniWords& dics, PRUnichar* current, int x, int y)
{
	miniHMenu hDics, hMenu;
	BOOL (*append_menu)(HMENU, UINT, UINT, PRUnichar*);
	BOOL (*append_menu_tr)(HMENU, UINT, UINT, char*);
	
	if (IsWindowUnicode(hwnd)) {
		append_menu = append_menuW;
		append_menu_tr = append_menuW_tr;
	} else {
		append_menu = append_menuA;
		append_menu_tr = append_menuA_tr;
	}
	// dictionaries is on sub menu.
	hDics = CreatePopupMenu();
	if (hDics == NULL) return -1;
	// append dictionaries
	for (int i = 0, id = 0x1000; i < dics.length(); ++ i, ++ id) {
		UINT flags = MF_STRING;
		// lstrcmpW can't compare correctly.
		flags |= wcscmp(current, dics[i]) ? MF_ENABLED : MF_CHECKED | MF_GRAYED;
		if (! append_menu(hDics, flags, id, dics[i])) {
			return -1;
		}
	}
	if (! append_menu(hDics, MF_SEPARATOR, 0, 0) ||
	    ! append_menu_tr(hDics, MF_ENABLED | MF_STRING, L_ID_SETDIC, LABEL_SETDIC)) {
		return -1;
	}
	// main popup menu
	hMenu = CreatePopupMenu();
	if (hMenu == NULL) return -1;
	// append suggested word
	for (int i = 0, id = 0x2000; i < suggest.length(); ++ i, ++ id) {
		if (! append_menu(hMenu, MF_ENABLED | MF_STRING, id, suggest[i])) {
			return -1;
		}
	}
	// append "add this word" and "ignore this word"
	if (suggest.length() > 0) {
		if (! append_menu(hMenu, MF_SEPARATOR, 0, 0) ||
		    ! append_menu_tr(hMenu, MF_ENABLED | MF_STRING, L_ID_IGNOREWORD, LABEL_IGNOREWORD) ||
		    ! append_menu(hMenu, MF_SEPARATOR, 0, 0) ||
		    ! append_menu_tr(hMenu, MF_ENABLED | MF_STRING, L_ID_ADDWORD, LABEL_ADDWORD)) {
			return -1;
		}
	}
	// append dictionary selecting popup
	if (! append_menu_tr(hMenu, MF_ENABLED | MF_POPUP | MF_STRING, (UINT) (HMENU) hDics, LABEL_SELECTDIC)) {
		return -1;
	}
	hDics.release();
	// append "disable spellchecker"
	if (! append_menu(hMenu, MF_SEPARATOR, 0, 0) ||
	    ! append_menu_tr(hMenu, MF_ENABLED | MF_STRING, L_ID_DISABLE, LABEL_DISABLE)) {
		return -1;
	}
	INT result = TrackPopupMenu(hMenu, TPM_TOPALIGN | TPM_LEFTALIGN | TPM_RETURNCMD | TPM_NONOTIFY, x, y, 0, hwnd, NULL);
	return result;
}


int message_box(HWND hwnd, LPCSTR lpText, LPCSTR lpTitle, UINT style)
{
	if (IsWindowUnicode(hwnd)) {
		return MessageBoxW(hwnd, miniWString(kPlugin.kFuncs->Translate(lpText)), miniWString(kPlugin.kFuncs->Translate(lpTitle)), style);
	}
	return MessageBoxA(hwnd, kPlugin.kFuncs->Translate(lpText), kPlugin.kFuncs->Translate(lpTitle), style);
}


/*
  Get screen position of a word by accesibility API
  provided mozilla and windows .
 */
BOOL get_word_pos(HWND hwnd, miniCOMPtr<nsIDOMNode>& node, PRInt32 offset, int& rx, int& ry)
{
	nsresult rv;
	
	miniCOMPtr<nsIAccessibleRetrieval> accessibleretrieval;
	rv = accessibleretrieval.doGetService("@mozilla.org/accessibleRetrieval;1");
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(accessibleretrieval, FALSE);
	
	miniCOMPtr<nsIAccessible> accessible;
	rv = accessibleretrieval->GetAccessibleFor(node, &accessible);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(accessible, FALSE);
	
	// below code is related with Microsoft Active Accessibility
	
	// This is MS COM component, isn't XPCOM.
	IAccessible* pAccessible = NULL;
	rv = accessible->GetNativeInterface((void**) &pAccessible);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(pAccessible, FALSE);
	
	HRESULT hresult;
	IServiceProvider *pServProv = NULL;
	hresult = pAccessible->QueryInterface(IID_IServiceProvider, (void**) &pServProv);
	pAccessible->Release();
	if (FAILED(hresult) || pServProv == NULL) return FALSE;
	
	ISimpleDOMText *pSimpleDOMText = NULL;
	hresult = pServProv->QueryService(IID_ISimpleDOMText, IID_ISimpleDOMText, (void**) &pSimpleDOMText);
	pServProv->Release();
	if (FAILED(hresult) || pSimpleDOMText == NULL) return FALSE;
	
	int x, y, w, h;
	hresult = pSimpleDOMText->get_clippedSubstringBounds(offset, offset + 1, &x, &y, &w, &h);
	pSimpleDOMText->Release();
	if (FAILED(hresult)) return FALSE;
	
	rx = x; ry = y + h;
	return TRUE;
}


/*
  Sort Suggestion List.
  
 */
VOID SortSuggestionList(miniWords& suggests)
{
	int order = 0;
	
	kPlugin.kFuncs->GetPreference(PREF_INT, PREFKEY_SUGGEST_ORDER, &order, &order);
	
	switch (order) {
	case 1:
		suggests.sort();
		break;
	default:
		break;
	}
}


int GetMaxSuggests()
{
	int n = 20;
	
	kPlugin.kFuncs->GetPreference(PREF_INT, PREFKEY_MAX_SUGGESTS, &n, &n);
	if (n > 30) {
		n = 30;
	}
	else if (n < 5) {
		n = 5;
	}
	return n;
}


/*
  Almost steps are borrowed from seamonkey.
  
  mozilla/editor/ui/composer/content/editorInlineSpellCheck.js
  mozilla/toolkit/content/inlineSpellCheckUI.js
 */

BOOL DoCommand(HWND hwnd, BOOL bHere)
{
	nsresult rv;
	nsEmbedString word;
	
	miniCOMPtr<nsIEditor> editor;
	if (! get_editor(hwnd, editor)) return FALSE;
	
	miniCOMPtr<nsIInlineSpellChecker> ispell;
	rv = editor->GetInlineSpellChecker(&ispell);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(ispell, FALSE);
	
	// Is spellchecker enabled?
	PRBool is_enabled = 0;
	rv = ispell->GetEnableRealTimeSpell(&is_enabled);
	NS_ENSURE_SUCCESS(rv, FALSE);
	if (! is_enabled) {
		int result = message_box(hwnd, LABEL_ENABLE, LABEL_TITLE, MB_YESNOCANCEL | MB_ICONQUESTION);
		if (result == IDYES) {
			// enable spellchecher
			rv = ispell->SetEnableRealTimeSpell(true);
			NS_ENSURE_SUCCESS(rv, FALSE);
		}
		// Can't continue. Must return.
		return TRUE;
	}
	
	miniCOMPtr<nsISelection> selection;
	rv = editor->GetSelection(&selection);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(selection, FALSE);
	
	miniCOMPtr<nsIDOMNode> node;
	rv = selection->GetAnchorNode(&node);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(node, FALSE);
	
	PRInt32 offset;
	rv = selection->GetAnchorOffset(&offset);
	NS_ENSURE_SUCCESS(rv, FALSE);
	
	miniCOMPtr<nsIEditorSpellCheck> spell;
	rv = ispell->GetSpellChecker(&spell);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(spell, FALSE);
	
	PRBool is_mis = 0;
	miniCOMPtr<nsIDOMRange> range;
	rv = ispell->GetMispelledWord(node, offset, &range);
	if (NS_SUCCEEDED(rv) && range) {
		rv = range->ToString(word);
		NS_ENSURE_SUCCESS(rv, FALSE);
		NS_ENSURE_TRUE(range, FALSE);
		
		rv = spell->CheckCurrentWord(word.get(), &is_mis);
		NS_ENSURE_SUCCESS(rv, FALSE);
	}
	
	// get suggestions list
	const int max_suggests = GetMaxSuggests();
	miniWords suggests(max_suggests);
	if (is_mis) {
		for (int i = 0; i < max_suggests; ++ i) {
			PRUnichar *w;
			rv = spell->GetSuggestedWord(&w);
			NS_ENSURE_SUCCESS(rv, FALSE);
			if (! w) break;
			if (! *w) {
				nsMemory::Free(w);
				break;
			}
			suggests.push(w);
		}
	}
	SortSuggestionList(suggests);
	
	// get dictionaries list
	PRUnichar** d;
	PRUint32 c;
	rv = spell->GetDictionaryList(&d, &c);
	NS_ENSURE_SUCCESS(rv, FALSE);
	miniWords dics(d, c);
	dics.sort();
	
	// get current dictionary
	miniPtr<PRUnichar> dict;
	rv = spell->GetCurrentDictionary(&dict);
	NS_ENSURE_SUCCESS(rv, FALSE);
	
	// get menu position
	int x, y;
	if (! bHere || ! get_word_pos(hwnd, node, offset, x, y)) {
		POINT point = {0};
		GetCursorPos(&point);
		x = point.x; y = point.y;
	}
	
	// show menu
	INT result = select_menu(hwnd, suggests, dics, dict, x, y);
	if (result < 1) {
		// canceled
	} else if (result == L_ID_DISABLE) {
		rv = ispell->SetEnableRealTimeSpell(false);
		NS_ENSURE_SUCCESS(rv, FALSE);
	} else if (result == L_ID_ADDWORD) {
		rv = ispell->AddWordToDictionary(word);
		NS_ENSURE_SUCCESS(rv, FALSE);
	} else if (result == L_ID_IGNOREWORD) {
		rv = ispell->IgnoreWord(word);
		NS_ENSURE_SUCCESS(rv, FALSE);
	} else if (result == L_ID_SETDIC) {
		// mozilla/editor/composer/src/nsEditorSpellCheck.cpp
		// SaveDefaultDictionary
		kPlugin.kFuncs->SetPreference(PREF_UNISTRING, PREFKEY_DICTIONARY, (PRUnichar*) dict, TRUE);
	} else if (result & 0x1000) {
		// change dictionary
		result &= 0xfff;
		if (result < dics.length()) {
			rv = spell->SetCurrentDictionary(dics[result]);
			NS_ENSURE_SUCCESS(rv, FALSE);
			// re-check all
			rv = ispell->SpellCheckRange(0);
			NS_ENSURE_SUCCESS(rv, FALSE);
		}
	} else if (result & 0x2000) {
		// correct word
		result &= 0xfff;
		if (result < suggests.length()) {
			word = suggests[result];
			rv = ispell->ReplaceWord(node, offset, word);
			NS_ENSURE_SUCCESS(rv, FALSE);
		}
	}
	
	return TRUE;
}
